from .fileinfo import FileInfo
from . import ifileinfo, filewriter, CsvFileInfo, utils
import numpy as np
import os
import soundfile as sf

class Merger():
    """
            Merger allows for creation of an empty soundfile to store multiple datasets easily.
            
            NOTE:
                W64 filetype is recommended, given it can store up to 18 exabytes of data.

            Args:
                channels: Number of channels the files parsed should have.
                path: the path where the new merger file should be created.
                samplerate: samplerate to write on the file.
                cutoff: how often should the file be written **NOT IMPLEMENTED**
                    (eg. for each 1024MB of data reached a new file is created)
                mode: either 'a' or 'w' if the file should be appended or truncated, respectively.
                    Default behavior: append
    """
    def __init__(self, channels, path, samplerate, cutoff = 1024, mode='a'):
        
        self._channels = channels
        self._init_samples()
        if(isinstance(samplerate,int)):
            self._samplerate = samplerate
        else:
            raise TypeError("Samplerate not integer")
        self.path = path

        #if the file exists already, then there's no need to create it(FileInfo)
        #if theres no file, we need to create it (array)
        self.file_exists = (os.path.isfile(self.path))

        True #checks if the file exists
        self.mode = mode

    def _init_samples(self):
        self._samples = np.empty(shape=(0,self._channels))

    def __enter__(self):
        return self

    def __exit__(self,type,value,traceback):
        pass

    def add(self, *files):
        """
            Adds new samples to the buffer array.

            When ``create_file`` method is executed the buffer gets emptied.

        Args:
            *files: An array, containing several pathes to files or IFileInfos specializations,
                although the latter is preferred.
        """
        tmp = None
        for f in files:
            if(issubclass(type(f),ifileinfo.IFileInfo)):
                tmp = f.getSamples()
            else:
                with FileInfo(f) as this:
                    tmp = this.getSamples()
            self._samples = np.append(self._samples,tmp,axis=0)
            #if(sys.getsizeof(self._samples) > (2**29)): escreve no ficheiro; para tal é necessário ter o ficheiro pré definido
            #adicionar 2**29(512MB) como cutoff

    def create_file(self, samplerate = None):
        """
            Creates a new file with the filename, converts based on extension given in ``new_filename``
            
            When executed the sample buffer will be emptied,
            so ``create_file`` should be executed frequently.

            Args:
                samplerate: Samplerate of the file.

        """
        if(len(self._samples) == 0):
            raise TypeError("No samples are queued!")
        if(samplerate == None):
            samplerate = self._samplerate
        path = utils.path_splitter(self.path)
        if(path['name'] == None):
            raise TypeError("It doesn't contain an extension")
        #tries to create the directory
        try: 
            import os
            os.mkdir(path['path'])
        except (OSError,FileNotFoundError):
            import errno
            if(e.errno != errno.EEXIST):
                raise #if there's an error thats not eexits(file/directory exists)
        #Appends or simply write over the file based on mode 
        if(self.mode == 'a' and self.file_exists == True):
            with sf.SoundFile(path['full_path'], mode = 'r+') as wfile:
                wfile.seek(0,sf.SEEK_END)
                wfile.write(self._samples)
            self._init_samples()

        else: 
            with CsvFileInfo(self._samples,self._samplerate) as infofile:
                self.file_exists = True
                self._init_samples()
                with filewriter.FileWriter(infofile) as fw:
                    fw.create_file(path['full_path'], self._samplerate)


import click
@click.command()
@click.argument(
    'input',
    type=click.Path(exists=True,resolve_path=True),
)
@click.argument(
    'output',
    type=click.Path(file_okay=False,dir_okay=False,resolve_path=True),
)
@click.option(
    '--resample',
    '-r',
    help='Value the dataset must be resampled to',
)
@click.option(
    '--fix_duration',
    '-f',
    help='Duration every dataset file should have.',
)
def cli(resample,fix_duration,input,output):
    #it also should normalize but in order to do that it must accept an array of values...
    import os
    if(os.path.isdir(input)):
        print(">pasta")
        pass
    else:
        print(">ficheiro")
        pass
        
    from .utils import is_number   
    if(resample is not None):
        if(not is_number(resample)):
            print("Invalid resample value!")
            return
        resample= float(resample)

    if(fix_duration is not None):
        if(not is_number(fix_duration)):
            print("Invalid fix_duration value!")
        fix_duration = float(fix_duration)
    from .fileinfo import FileInfo
    from .fileutil import FileUtil
    from .filewriter import FileWriter
    print("Processing...")
    with FileInfo(input) as file:
        with FileUtil(file) as futil:
            if(fix_duration):
                futil.fix_duration(fix_duration)
            if(resample):
                futil.resample(resample)
        with FileWriter(file) as fw:
            fw.create_file(output)

cli()


from abc import ABC, abstractmethod

class IFileInfo(ABC):
    """
        Interface which must be implemented if you want to support your own filetype;
        Used as an argument to FileWriter, FileUtil and FileMerger.
    """
    def __init__(self,file):
        self.file = file
        pass

    @abstractmethod
    def __enter__(self):
        """
            When entering a context manager the enter method is called.
            This is the most correct way to manipulate files.
            Check the ``FileInfo`` class for examples.
        """

        pass
    @abstractmethod
    def __exit__(self,type,value,traceback):
        """
            When closing a context manager the exit method is called.
            This is the most correct way to manipulate files.
            Check the ``FileInfo`` class for examples.
        """

        pass

    @abstractmethod
    def getFilepath(self):
        """
            Returns:
                The filepath of the current file.
        """
        pass

    @abstractmethod
    def getNumberOfFrames(self):
        """
            Returns:
                The number of frames the file has.
        """
        pass

    @abstractmethod
    def getSamplerate(self):
        """
            Returns:
                The samplerate of the file.
        """
        pass

    @abstractmethod
    def getNumberOfChannels(self):
        """
            Returns:
                The number of channels the file has.
        """
        pass    

    @abstractmethod        
    def getSamples(self):
        """
            Returns:
                All the samples the file has.
        """
        pass

    # writing methods
    @abstractmethod
    def setSamples(self, samples):
        """
            Writes to the file the samples given in ``samples``.

            It will truncate the old samples. 
            If you want to add, use the ``addSamples`` method.

            Args:
                samples: An array containing the new samples.
                    It must be shaped like (n,c) where c is the number of channels
        """
        pass
    
    @abstractmethod
    def truncate(self, num_frames):
        """
            Truncates the file to have only the first ``num_frames`` samples.

            Args:
                num_frames: The number of frames the file will have.
        """

        pass

    @abstractmethod
    def addSamples(self, samples):
        """
            Adds samples given by ``samples``.

            Args: 
                samples: An array containing samples.
                    It must be shaped like (n,c) where c is the number of channels.
        """

        pass
    
    @abstractmethod
    def close(self):
        """
            Defines the behavior the class should have when leaves the context manager.
            
            If a file descriptor is being used you should always define the close method.
        """
        pass
    



def path_splitter(path):
    """
        Cleans extra / characters, 
        splits the path in 4 parts: See example

        Parameters:
            path:
                Receives a path

        Returns:
            tuple: dictionary with the following keys:
                {full_path, path, name, extension}

        Example:
            >>> path.splitter('C:/Data/example.wav/')
            {
                'full_path':'C:/Data/example.wav',
                'path':'C:/Data/',
                'name':'example.wav',
                'extension':'wav'
            }
    """

    if(path.endswith("/")): 
        #if it ends with a slash removes it
        path = path[:-1]
    stripped_path = path.split("/")
    if("." in path[2:]): #if it has a dot means it has an extension; [2:] is used in case it is a relative path
        path = '/'.join(stripped_path[:-1])+'/' #constructs the path, except the filename
        file_name = stripped_path[-1] #gets the filename
        extension = file_name.split('.')[-1] # gets the extension
    else:
        path = '/'.join(stripped_path)+'/' #constructs the path, except the filename
        file_name = None
        extension = None
    
    if(len(path) < 2):
        #if it has only a character(possibly /, so removes it in order to avoid refering to root on mac/linux)
        #now theres a problem... if the user really wants to refer to the root? Shame, you cannot lol
        path = path[:-1]
    return {'full_path': path+file_name ,'path': path,'name': file_name,'extension': extension}


def is_number(input):
    """
        Receives an input and checks if it is actually a number
    """
    try:
        float(input)
    except ValueError as id:
        return False
    else:
        return True


from .utils import path_splitter
import os

class FileWriter:
    """
        Writes to a file.

        The class should be instantiated with a ``with`` statement.

        Args:
            file: Accepts either a FileUtil, or IFileInfo Specialization.
            mode: Allows for `w` for writing or `a` for appending.
    """
    def __init__(self,file, mode = 'w'):
        from .ifileinfo import IFileInfo
        from .fileutil import FileUtil
        if(issubclass(type(file),IFileInfo)):
            self.file = file 
        elif(issubclass(type(file),FileUtil)):
            self.file = file.file
        else:
            raise TypeError("Wrong argument type")
        
        self.mode = mode

    def __enter__(self):
        return self

    def __exit__(self,type,value,traceback):
        self.close()
    def close(self):
        #self.file.close()
        pass
        
    def create_file(self,new_filepath,samplerate = None):
        """
            Creates a new file with the extension given in ``new_filepath``.

            If the source file isn't a soundfile the target file will not be well formated.

            In order to normalize, you should run ``FileUtil.standardize`` method before.

            Args:
                new_filepath - The diretory and name the new file will have,
                    it will convert based on file extension.
                samplerate (Optional) - The samplerate the file should have,
                    if not supplied it will use the own `file` samplerate.
        """

        file_exists = (os.path.isfile(new_filepath))
        if(samplerate == None):
            samplerate = self.file.getSamplerate()
        path = path_splitter(new_filepath)
        if(path['name'] == None):
            raise TypeError("It doesn't contain a filename")
        try:
            os.mkdir(path['path'])
            #tries to create the directory
        except (OSError,FileNotFoundError) as e:
            pass
        import soundfile as sf
        if(self.mode == 'a' and file_exists):
            with sf.SoundFile(path['full_path'], mode = 'r+') as wfile:
                wfile.seek(0,sf.SEEK_END)
                wfile.write(self.file.getSamples())
        else:
            sf.write(path['full_path'], self.file.getSamples(), samplerate,format=path['extension']) # writes to the new file 
        return


import soundfile as sf
import tempfile
import os
from .ifileinfo import IFileInfo

class FileInfo(IFileInfo):
    """
        Defines the class to manipulate soundfiles.


        Receives a path to a file.

        Copies the file to a temporary location.

        Gets edited through the FileUtil.

        FileWriter converts and writes to another location.


        NOTE:
            - close method MUST always be called or else the temporary file stays in disk.
            - ``with`` statements should be used in order to close the files automatically.

    """
    def __init__(self,path):
        super().__init__(path)
        self._path = path #ACTUAL FILE
        self._tmpPath = tempfile.gettempdir() #TEMPORARY COPY

        try:
            import shutil
            self._tmpPath = shutil.copy(self._path,self._tmpPath)
        except FileExistsError:
            print("FEXISTS l28")
            pass
        self._initinfo()

    def __enter__(self):
        return self

    def __exit__(self,type,value,traceback):
        self.close()

    def _initinfo(self):
        with sf.SoundFile(self._tmpPath, mode ='r') as rfile:
            self._samplerate = rfile.samplerate
            self._frames = rfile.read()
            self._numchannels = rfile.channels
        return

    #================READ METHODS ==================#

    def getFilepath(self):
        """
            See the base class ``ifileinfo``.
        """
        return self._path

    def getNumberOfFrames(self):
        """
            See the base class ``ifileinfo``.
        """
        return len(self.getSamples())
            
    def getSamplerate(self):
        """
            See the base class ``ifileinfo``.
        """
        return self._samplerate

    def getNumberOfChannels(self):
        """
        Returns:
            Number of channels the file has.
        """
        return self._numchannels

    def getDuration(self):
        """
        Returns:
            The duration of the file in seconds.
        """
        return self.getNumberOfFrames()/self.getSamplerate()

    def get_rounded_duration(self):
        """
            Returns:
                The rounded duration in seconds.
        """
        from math import ceil
        rounded_duration = ceil(self.getDuration()) #in seconds
        return rounded_duration 
    
    def getSamples(self):
        """
        Reads all of the samples in the file

        Returns: 
            numpy array containing the samples.
        """
        return self._frames
        
    #================WRITE METHODS ==================#

    def setSamples(self, samples, samplerate = None):
        """
            Writes the ´samples´ as the new samples in the file.

            Args:
                samples: numpy array shaped like (n,c), where c is the number of channels.
                samplerate: the new sample rate the file will have, if none it will use the initial samplerate.
        """ 
        if (samplerate == None):
            samplerate = self.getSamplerate()
        samples = self._soundfile_array_converter(samples)
        with sf.SoundFile(self._tmpPath, mode = 'w', samplerate = self.getSamplerate(), channels = self.getNumberOfChannels()) as wfile:
            wfile.truncate(0) #removes all the samples in the file                
        sf.write(self._tmpPath,samples,samplerate)
        self._initinfo()

    def truncate(self,num_frames):
        """
            Truncates the file to only have ``num_frames``.
        """
        #with sf.SoundFile(self._tmpPath, mode = 'w', samplerate = self.getSamplerate(), channels = self.getNumChannels()) as wfile:
        #   wfile.truncate(num_frames)
        self.setSamples(self.getSamples()[:num_frames])
        self._initinfo()

    def addSamples(self,samples):
        """
            Appends the samples to the file.

            Similar to ´setSamples()´ but appends instead of truncating.
            See the base class ``ifileinfo``.

        """ 

        samples = self._soundfile_array_converter(samples)
        with sf.SoundFile(self._tmpPath, mode = 'r+') as wfile:
            wfile.seek(0,whence = sf.SEEK_END)
            wfile.write(samples)

    #================MISC METHODS ==================#
 
    def _soundfile_array_converter(self,array):
        """
            Checks if it has the correct array shape to be worked by pysoundfile.
            If not, it will transpose de array.
            Arguments:
                array: ndarray to convert.
            Returns:
                array: the converted array.
        """
        if(array.shape[1] > 20): #meaning it doesnt represent the number of channelsnumber of channels
            array = array.T
        return array
    
    def close(self):
        """
            Must always be called or the file won't be accessible by other processes AND the temp file will stay in disk.
            See the base class ``ifileinfo``.

        """
        import os
        try:
            os.unlink(self._tmpPath)

        except expression as identifier:
            print("err", identifier)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
with open("requirements.txt","r") as req:
    inst_req = req.read()
setuptools.setup(
    name='dscleaner',
    version='0.1.8.post2',
    author='Manuel Pereira',
    author_email='afonso.pereira4525@gmail.com',
    packages=['dscleaner',],
    project_urls={
        "Documentation": "https://dscleaner.rtfd.io/",
        "Source code": "https://gitlab.com/ManelPereira/dscleaner",
    },
    license='The MIT License',
    description='Removed debugging prints and fixed resampled method on cli.',
    long_description=long_description,
    install_requires=inst_req,
)


import os
import numpy as np
from math import ceil
class FileUtil():
    """
        FileUtil class is where the dataset manipulation occur.
        
        The class should be instantiated with a ``with`` statement.

        Args:
            f: a IFileInfo specialization must be supplied!
    """
    def __init__(self, f):
        
        from .ifileinfo import IFileInfo
        assert issubclass(type(f),IFileInfo),"Not a specialization of IFileInfo"
        self.file = f
    
    def __enter__(self):
        return self

    def __exit__(self,type,value,traceback):
        pass

    def fix_duration(self, expected_duration, grid_rate = 50):
        """
            Fixes the file to the expected duration.
            
            Args:
                expected_duration: Duration the file should have in minutes.
                grid_rate: frequency of the grid in hertz, this is used to discover the wave signal in order to upsample.
        """
        data = self.file.getSamples()
        sample_rate = self.file.getSamplerate()

        actual_frame_number = len(data)
        expected_frame_number = ceil(sample_rate * expected_duration * 60)
        #rate * minutes * seconds
        if(actual_frame_number == expected_frame_number):
            #no cleaning needed
            return
        else:
            if(actual_frame_number > expected_frame_number):
                self.file.truncate(expected_frame_number)
            else:
                #needs adding
                wave_length = int((sample_rate / grid_rate) )
                last_wave = list(data[-wave_length:]) # gets the last wave_length elements
                samples_missing = int(expected_frame_number - actual_frame_number)
                last_wave *= ceil(samples_missing/wave_length) #replicates the last wave several times
                last_wave = np.asarray(last_wave[:samples_missing]) #removes any more than its actually missing
                self.file.addSamples(last_wave)
    
    def standardize(self,*dividers):
        """
            This method transforms the values to fit between -1 and 1, in order to be used in soundfiles.

            If the source file isn't a soundfile the target file will not be well formated,
            hence you should run this method to make the file well formated.
            
            Args:
                *dividers: The number which each channel will be divided by in order to standardize that channel.

            NOTE:
                In order to maintain consistency throughout the dataset it is advised that the
                divider chosen for each channel to be a bit higher than the max value.
                It is also advised to keep record of the divider for each channel for future unstardartization.
                    
                Example:
                    Max amplitude is 75
                    divider chosen: 90.

            Returns:
                A tuple with the dividers used to standardize.
                    Example:
                        (40,30,30) in a three channel file.
                        
                You should keep these values for future reference.
             
        """
        iter = 0
        print(type(dividers))
        dividers = list(dividers) #convert tuple to list
        if(len(dividers)>self.file.getNumberOfChannels()):
            raise TypeError("There are more dividers than channels in the file!")
        elif(len(dividers) == 0): #if there are no dividers we find the max value in each channel and divide by it
            for i in range(0,self.file.getNumberOfChannels()):
                try:
                    dividers.append(ceil(max(self.file.getSamples()[:,i]))) #iterates over each channel
                except IndexError as e:
                    dividers.append(ceil(max(self.file.getSamples())))
        try:
            tmp = np.ndarray(shape = self.file.getSamples().shape)
        except IndexError as e:
            tmp = np.array()
        for divider in dividers:
            if(divider in (int,float,0)):
                TypeError("Not a number")
            try:
                tmp[:,iter] = self.file.getSamples()[:,iter]/divider
            except IndexError as e:
                tmp = self.file.getSamples()/divider   
            iter += 1
        self.file.setSamples(tmp)
        return (dividers)

    def resample(self, new_framerate, method = 'kaiser_fast'):
        """
            Resamples the data to the new framerate using librosa resample.
            
            Args:
                data: numpy.array shaped like (num_frames,num_channels) is expected to receive the soundfile.getSamples()
                    not the transposed array.
                original_framerate: the original framerate the data array uses.
                new_framerate: the new framerate that data will be resampled to.
                method: Methods that librosa accepts are also accepted here, uses `kaiser_fast` by default.
        """
        data = self.file.getSamples()
        new_framerate = ceil(new_framerate)
        import librosa
        new_data = librosa.resample(data.T, self.file.getSamplerate(), new_framerate, res_type=method, fix=True)
        self.file.setSamples(new_data.T,new_framerate)


from . import ifileinfo
import pandas as pd

class CsvFileInfo(ifileinfo.IFileInfo):
    """
        CsvFileInfo is used when there is no actual file, but an array.
        
        The array must be shaped in (n,c) where c is the number of channels.
    """
    def __init__(self, samples, samplerate):
        self._samples = samples
        self._samplerate = samplerate
        pass

    def __enter__(self):
        """
            See the base class ``ifileinfo``.
        """
        return self

    def __exit__(self,type,value,traceback):
        """
            See the base class ``ifileinfo``.
        """
        self.close()

    def getSamples(self):
        """
            See the base class ``ifileinfo``.
        """

        return self._samples

    def getSamplerate(self):
        """
            See the base class ``ifileinfo``.
        """

        return self._samplerate
    
    def getNumberOfFrames(self):
        """
            See the base class ``ifileinfo``.
        """

        return len(self._samples)
    
    def getNumberOfChannels(self):
        """
            See the base class ``ifileinfo``.
        """

        try:
            numchannels = self._samples.shape[1]
        except IndexError as e:
            numchannels = 1
        return numchannels

    def getFilepath(self):
        """
            Don't rely on this method because it's not implemented.
        """
        raise NotImplementedError("Not avaliable in csvFileInfo")

    #WRITE
    def setSamples(self, samples, framerate = None):
        """
            See the base class ``ifileinfo``.
        """

        if (framerate == None):
            framerate = self.getSamplerate()
        self._samples = samples
        self._samplerate = framerate
        return
    
    def truncate(self, num_frames):
        """
            See the base class ``ifileinfo``.
        """

        self._samples = self._samples[:num_frames]
        return

    def addSamples(self, samples):
        """
            See the base class ``ifileinfo``.
        """

        self._samples.append(samples)
        return 

    def close(self):
        """
            Does nothing since the sample container is an array.
        """
        
        return